﻿#include "pch.h"
#include "MeshTestRenderer.h"

#include "Common/DirectXHelper.h"

using namespace MeshTest;

using namespace D2D1;
using namespace Microsoft::WRL;

MeshTestRenderer::MeshTestRenderer(const std::shared_ptr<DX::DeviceResources>& deviceResources) : 
    m_text(L""),
    m_deviceResources(deviceResources),
    m_renderingOption(RenderingOption::None),
    m_rotationAngle(0)
{
    // Create device independent resources
    DX::ThrowIfFailed(
        m_deviceResources->GetDWriteFactory()->CreateTextFormat(
            L"Segoe UI",
            nullptr,
            DWRITE_FONT_WEIGHT_LIGHT,
            DWRITE_FONT_STYLE_NORMAL,
            DWRITE_FONT_STRETCH_NORMAL,
            32.0f,
            L"en-US",
            &m_textFormat
            )
        );

    DX::ThrowIfFailed(
        m_textFormat->SetParagraphAlignment(DWRITE_PARAGRAPH_ALIGNMENT_FAR)
        );

    DX::ThrowIfFailed(
        m_textFormat->SetTextAlignment(DWRITE_TEXT_ALIGNMENT_TRAILING)
        );

    DX::ThrowIfFailed(
        m_deviceResources->GetD2DFactory()->CreateDrawingStateBlock(&m_stateBlock)
        );

    CreateDeviceDependentResources();
}

void MeshTestRenderer::CreateDeviceDependentResources()
{
    DX::ThrowIfFailed(
        m_deviceResources->GetD2DDeviceContext()->CreateSolidColorBrush(D2D1::ColorF(D2D1::ColorF::Blue), &m_blueBrush)
        );
}

void MeshTestRenderer::ReleaseDeviceDependentResources()
{
    m_blueBrush.Reset();
    m_starMesh.Reset();
    m_realizedStarGeometry.Reset();
}

void MeshTestRenderer::CreateWindowSizeDependentResources()
{
    Windows::Foundation::Size logicalSize = m_deviceResources->GetLogicalSize();

    // Create 201-point star
    const float pi = 3.14159f;
    const int numVertices = 201;        // odd number!
    const float increment = pi * (numVertices - 1) / numVertices;
    const float radius = min(logicalSize.Width, logicalSize.Height) / 2 - 50;
    float radians = 0;

    DX::ThrowIfFailed(
        m_deviceResources->GetD2DFactory()->CreatePathGeometry(&m_starGeometry)
        );

    ComPtr<ID2D1GeometrySink> geometrySink;

    DX::ThrowIfFailed(
        m_starGeometry->Open(&geometrySink)
        );

    for (int i = 0; i < numVertices; i++)
    {
        float x = radius * sinf(radians);
        float y = -radius * cosf(radians);

        if (i == 0)
            geometrySink->BeginFigure(Point2F(x, y), D2D1_FIGURE_BEGIN_FILLED);
        else
            geometrySink->AddLine(Point2F(x, y));

        radians += increment;
        radians = fmodf(radians, 2 * pi);
    }

    geometrySink->EndFigure(D2D1_FIGURE_END_CLOSED);

    DX::ThrowIfFailed(
        geometrySink->Close()
        );

    // Create outlined path geometry
    DX::ThrowIfFailed(
        m_deviceResources->GetD2DFactory()->CreatePathGeometry(&m_outlinedStarGeometry)
        );

    DX::ThrowIfFailed(
        m_outlinedStarGeometry->Open(&geometrySink)
        );

    DX::ThrowIfFailed(
        m_starGeometry->Outline(Matrix3x2F::Identity(), geometrySink.Get())
        );

    DX::ThrowIfFailed(
        geometrySink->Close()
        );

    // Create flattened path geometry
    DX::ThrowIfFailed(
        m_deviceResources->GetD2DFactory()->CreatePathGeometry(&m_flattenedStarGeometry)
        );

    DX::ThrowIfFailed(
        m_flattenedStarGeometry->Open(&geometrySink)
        );

    DX::ThrowIfFailed(
        m_starGeometry->Simplify(D2D1_GEOMETRY_SIMPLIFICATION_OPTION_LINES, Matrix3x2F::Identity(), geometrySink.Get())
        );

    DX::ThrowIfFailed(
        geometrySink->Close()
        );

    // Create mesh
    DX::ThrowIfFailed(
        m_deviceResources->GetD2DDeviceContext()->CreateMesh(&m_starMesh)
        );

    ComPtr<ID2D1TessellationSink> tessellationSink;

    DX::ThrowIfFailed(
        m_starMesh->Open(&tessellationSink)
        );

    DX::ThrowIfFailed(
        m_starGeometry->Tessellate(Matrix3x2F::Identity(), tessellationSink.Get())
        );

    DX::ThrowIfFailed(
        tessellationSink->Close()
        );

    // Create geometry realization
    m_deviceResources->GetD2DDeviceContext()->CreateFilledGeometryRealization(m_starGeometry.Get(), 0.25f, &m_realizedStarGeometry);
}

void MeshTestRenderer::SetRenderingOption(RenderingOption renderingOption)
{
    m_renderingOption = renderingOption;
}

void MeshTestRenderer::Update(DX::StepTimer const& timer)
{
    // Update FPS display text.
    uint32 fps = timer.GetFramesPerSecond();
    m_text = (fps > 0) ? std::to_wstring(fps) + L" FPS" : L" - FPS";

    // Update rotation angle
    m_rotationAngle += float(360 * timer.GetElapsedSeconds() / 36);
    m_rotationAngle = fmodf(m_rotationAngle, 360);
}

// Renders a frame to the screen.
void MeshTestRenderer::Render()
{
    if (m_starGeometry == nullptr)
        return;

    ID2D1DeviceContext1* context = m_deviceResources->GetD2DDeviceContext();
    Windows::Foundation::Size logicalSize = m_deviceResources->GetLogicalSize();

    context->SaveDrawingState(m_stateBlock.Get());
    context->BeginDraw();
    context->Clear(ColorF(ColorF::LightYellow));

    // Display FPS text
    D2D1_RECT_F layoutRect = RectF(0, 0, logicalSize.Width, logicalSize.Height);
    context->SetTransform(m_deviceResources->GetOrientationTransform2D());
    context->DrawText(m_text.c_str(), m_text.length(), m_textFormat.Get(), &layoutRect, m_blueBrush.Get());

    // Display geometry (or mesh)
    Matrix3x2F center = Matrix3x2F::Translation(logicalSize.Width / 2, logicalSize.Height / 2);
    Matrix3x2F rotate = Matrix3x2F::Rotation(m_rotationAngle);
    context->SetTransform(rotate * center * m_deviceResources->GetOrientationTransform2D());

    switch (m_renderingOption)
    {
    case RenderingOption::None :
        break;

    case RenderingOption::PathGeometry:
        context->FillGeometry(m_starGeometry.Get(), m_blueBrush.Get());
        break;

    case RenderingOption::OutlinedPathGeometry:
        context->FillGeometry(m_outlinedStarGeometry.Get(), m_blueBrush.Get());
        break;

    case RenderingOption::FlattenedPathGeometry:
        context->FillGeometry(m_flattenedStarGeometry.Get(), m_blueBrush.Get());
        break;

    case RenderingOption::Mesh:
        context->SetAntialiasMode(D2D1_ANTIALIAS_MODE_ALIASED);
        context->FillMesh(m_starMesh.Get(), m_blueBrush.Get());
        break;

    case RenderingOption::GeometryRealization:
        context->DrawGeometryRealization(m_realizedStarGeometry.Get(), m_blueBrush.Get());
        break;
    }

    // Ignore D2DERR_RECREATE_TARGET here. This error indicates that the device
    // is lost. It will be handled during the next call to Present.
    HRESULT hr = context->EndDraw();
    if (hr != D2DERR_RECREATE_TARGET)
    {
        DX::ThrowIfFailed(hr);
    }

    context->RestoreDrawingState(m_stateBlock.Get());
}

