﻿#include "pch.h"
#include "OutThereAndBackAgainRenderer.h"
#include "Common/DirectXHelper.h"
#include "InterrogableTessellationSink.h"

using namespace OutThereAndBackAgain;

using namespace D2D1;
using namespace Microsoft::WRL;
using namespace Platform;
using namespace Windows::ApplicationModel;

OutThereAndBackAgainRenderer::OutThereAndBackAgainRenderer(const std::shared_ptr<DX::DeviceResources>& deviceResources) : 
    m_text(L""),
    m_deviceResources(deviceResources)
{
//	ZeroMemory(&m_textMetrics, sizeof(DWRITE_TEXT_METRICS));

    // Create device independent resources
    DX::ThrowIfFailed(
        m_deviceResources->GetDWriteFactory()->CreateTextFormat(
            L"Segoe UI",
            nullptr,
            DWRITE_FONT_WEIGHT_LIGHT,
            DWRITE_FONT_STYLE_NORMAL,
            DWRITE_FONT_STRETCH_NORMAL,
            32.0f,
            L"en-US",
            &m_textFormat
            )
        );

    DX::ThrowIfFailed(
        m_textFormat->SetParagraphAlignment(DWRITE_PARAGRAPH_ALIGNMENT_FAR)
        );

    DX::ThrowIfFailed(
        m_textFormat->SetTextAlignment(DWRITE_TEXT_ALIGNMENT_TRAILING)
        );

    DX::ThrowIfFailed(
        m_deviceResources->GetD2DFactory()->CreateDrawingStateBlock(&m_stateBlock)
        );

    CreateDeviceDependentResources();
}

void OutThereAndBackAgainRenderer::CreateDeviceDependentResources()
{
    ID2D1DeviceContext* context = m_deviceResources->GetD2DDeviceContext();

    DX::ThrowIfFailed(
        context->CreateSolidColorBrush(ColorF(ColorF::Blue), &m_blueBrush)
        );

    // Create gradient brush for text triangle rendering
    D2D1_GRADIENT_STOP gradientStops [] =
    {
        GradientStop(0, ColorF(ColorF::Red)),
        GradientStop(1, ColorF(ColorF::Blue))
    };

    ComPtr<ID2D1GradientStopCollection> gradientStopCollection;

    DX::ThrowIfFailed(
        context->CreateGradientStopCollection(gradientStops, ARRAYSIZE(gradientStops), &gradientStopCollection)
        );

    DX::ThrowIfFailed(
        context->CreateLinearGradientBrush(LinearGradientBrushProperties(Point2F(), Point2F()), 
                                           gradientStopCollection.Get(), 
                                           &m_gradientBrush)
        );
}

void OutThereAndBackAgainRenderer::ReleaseDeviceDependentResources()
{
    m_blueBrush.Reset();
    m_gradientBrush.Reset();
}

void OutThereAndBackAgainRenderer::CreateWindowSizeDependentResources()
{
    Windows::Foundation::Size logicalSize = m_deviceResources->GetLogicalSize();

    if (logicalSize.IsEmpty)
        return;

    m_triangleInfos.clear();

    // Load font file
    ComPtr<IDWriteFactory> dwriteFactory = m_deviceResources->GetDWriteFactory();
    String^ filePath = Package::Current->InstalledLocation->Path + "\\Fonts\\Pesca.ttf";
    ComPtr<IDWriteFontFile> fontFile;

    DX::ThrowIfFailed(
        dwriteFactory->CreateFontFileReference(filePath->Data(), nullptr, &fontFile)
        );

    // Create font face
    ComPtr<IDWriteFontFace> fontFace;

    DX::ThrowIfFailed(
        dwriteFactory->CreateFontFace(DWRITE_FONT_FACE_TYPE_TRUETYPE,
        1,
        fontFile.GetAddressOf(),
        0,
        DWRITE_FONT_SIMULATIONS_NONE,
        &fontFace)
        );

    // Obtain glyph geometry for "DirectX"
    UINT16 glyphIndices[] = { 0xC3, 0x4C, 0x55, 0x48, 0x46, 0x57, 0x3B };

    ComPtr<ID2D1Factory2> d2dFactory = m_deviceResources->GetD2DFactory();
    ComPtr<ID2D1PathGeometry> pathGeometry;

    DX::ThrowIfFailed(
        d2dFactory->CreatePathGeometry(&pathGeometry)
        );

    ComPtr<ID2D1GeometrySink> geometrySink;

    DX::ThrowIfFailed(
        pathGeometry->Open(&geometrySink)
        );

    float emSize = logicalSize.Width / 5;

    DX::ThrowIfFailed(
        fontFace->GetGlyphRunOutline(emSize,
                                     glyphIndices,
                                     nullptr,
                                     nullptr,
                                     ARRAYSIZE(glyphIndices),
                                     false,
                                     false,
                                     geometrySink.Get())
        );

    DX::ThrowIfFailed(
        geometrySink->Close()
        );

    // Get the geometry bounds for rendering purposes
    DX::ThrowIfFailed(
        pathGeometry->GetBounds(IdentityMatrix(), &m_geometryBounds)
        );

    // Tessellate geometry into triangles
    ComPtr<InterrogableTessellationSink> tessellationSink = new InterrogableTessellationSink();

    DX::ThrowIfFailed(
        pathGeometry->Tessellate(IdentityMatrix(), tessellationSink.Get())
        );

    std::vector<D2D1_TRIANGLE> triangles = tessellationSink->GetTriangles();

    for (D2D1_TRIANGLE triangle : triangles)
    {
        // Create a mesh for each triangle
        ComPtr<ID2D1Mesh> triangleMesh;

        DX::ThrowIfFailed(
            m_deviceResources->GetD2DDeviceContext()->CreateMesh(&triangleMesh)
            );

        ComPtr<ID2D1TessellationSink> sink;

        DX::ThrowIfFailed(
            triangleMesh->Open(&sink)
            );

        sink->AddTriangles(&triangle, 1);

        DX::ThrowIfFailed(
            sink->Close()
            );

        // Create a TriangleInfo object for each triangle
        TriangleInfo triangleInfo;
        triangleInfo.triangle = triangle;   // save for "single mesh" rendering
        triangleInfo.triangleMesh = triangleMesh;

        // Compute geometric center of triangle for rotations
        triangleInfo.triangleCenter = 
            Point2F((triangle.point1.x + triangle.point2.x + triangle.point3.x) / 3,
                    (triangle.point1.y + triangle.point2.y + triangle.point3.y) / 3);

        // Construct "travel" geometry
        ComPtr<ID2D1PathGeometry> travelGeometry;

        DX::ThrowIfFailed(
            d2dFactory->CreatePathGeometry(&travelGeometry)
            );

        DX::ThrowIfFailed(
            travelGeometry->Open(&geometrySink)
            );

        std::function<float()> RandomValue = [emSize]()
        {
            float range = 2 * emSize;
            return fmodf((float) rand(), range) - range / 2;
        };

        std::function<D2D1_POINT_2F()> RandomPoint = [RandomValue]()
        {
            return Point2F(RandomValue(), RandomValue());
        };

        geometrySink->BeginFigure(Point2F(), D2D1_FIGURE_BEGIN_HOLLOW);
        geometrySink->AddQuadraticBezier(QuadraticBezierSegment(RandomPoint(), RandomPoint()));
        geometrySink->EndFigure(D2D1_FIGURE_END_OPEN);

        DX::ThrowIfFailed(
            geometrySink->Close()
            );

        triangleInfo.travelGeometry = travelGeometry;

        // Get information about travel geometry
        DX::ThrowIfFailed(
            travelGeometry->ComputeLength(IdentityMatrix(), &triangleInfo.travelLength)
            );

        D2D1_POINT_2F point;    // will equal Point2F(0, 0) here
        D2D1_POINT_2F vector;

        DX::ThrowIfFailed(
            travelGeometry->ComputePointAtLength(0, IdentityMatrix(), &point, &vector)
            );

        triangleInfo.initialAngle = atan2(vector.y, vector.x);

        // Save TriangleInfo object in vector
        m_triangleInfos.push_back(triangleInfo);
    }

    // Create equal-sized vector for D2D1_TRIANGLES for single-FillMesh mode
    m_triangles = std::vector<D2D1_TRIANGLE>(m_triangleInfos.size());

    // Set start and end point for linear gradient brush
    m_gradientBrush->SetStartPoint(Point2F(m_geometryBounds.left, m_geometryBounds.top));
    m_gradientBrush->SetEndPoint(Point2F(m_geometryBounds.right, m_geometryBounds.bottom));
}

void OutThereAndBackAgainRenderer::SetRenderMode(bool useSingleFillMesh)
{
    m_useSingleFillMesh = useSingleFillMesh;
}

void OutThereAndBackAgainRenderer::Update(DX::StepTimer const& timer)
{
    // Update FPS display text.
    uint32 fps = timer.GetFramesPerSecond();
    m_text = (fps > 0) ? std::to_wstring(fps) + L" FPS" : L" - FPS";

    if (m_triangleInfos.size() == 0)
        return;

    // Calculate an interpolation factor
    float t = (float) fmod(timer.GetTotalSeconds() + 5, 10) / 10;
    float radians = 2 * 3.14159f * t;
    float fraction = max(0, std::sin(radians));

    for (unsigned int index = 0; index < m_triangleInfos.size(); index++)
    {
        TriangleInfo& triangleInfo = m_triangleInfos.at(index);
        float length = fraction * triangleInfo.travelLength;

        D2D1_POINT_2F point;
        D2D1_POINT_2F vector;

        DX::ThrowIfFailed(
            triangleInfo.travelGeometry->ComputePointAtLength(length, IdentityMatrix(), &point, &vector)
            );

        float angle = atan2(vector.y, vector.x);
        float degreeDiff = 180 * (angle - triangleInfo.initialAngle) / 3.14159f;

        Matrix3x2F matrix = 
            Matrix3x2F::Rotation(degreeDiff, triangleInfo.triangleCenter) * 
            Matrix3x2F::Translation(point.x, point.y);

        triangleInfo.matrix = matrix;

        // Accumulate transformed triangles in an array
        if (m_useSingleFillMesh)
        {
            D2D1_TRIANGLE triangle;
            triangle.point1 = matrix.TransformPoint(triangleInfo.triangle.point1);
            triangle.point2 = matrix.TransformPoint(triangleInfo.triangle.point2);
            triangle.point3 = matrix.TransformPoint(triangleInfo.triangle.point3);
            m_triangles.at(index) = triangle;
        }
    }

    // Create a single mesh from the triangles
    if (m_useSingleFillMesh)
    {
        DX::ThrowIfFailed(
            m_deviceResources->GetD2DDeviceContext()->CreateMesh(&m_textMesh)
            );

        ComPtr<ID2D1TessellationSink> sink;

        DX::ThrowIfFailed(
            m_textMesh->Open(&sink)
            );

        sink->AddTriangles(m_triangles.data(), m_triangles.size());

        DX::ThrowIfFailed(
            sink->Close()
            );
    }
}

// Renders a frame to the screen.
void OutThereAndBackAgainRenderer::Render()
{
    ID2D1DeviceContext* context = m_deviceResources->GetD2DDeviceContext();
    Windows::Foundation::Size logicalSize = m_deviceResources->GetLogicalSize();

    context->SaveDrawingState(m_stateBlock.Get());
    context->BeginDraw();
    context->Clear(ColorF(ColorF::LightCyan));

    // Display FPS text
    context->SetTransform(m_deviceResources->GetOrientationTransform2D());
    context->DrawText(m_text.c_str(), 
                      m_text.length(), 
                      m_textFormat.Get(), 
                      RectF(0, 0, logicalSize.Width, logicalSize.Height), 
                      m_blueBrush.Get());

    // Display meshes
    Matrix3x2F centerMatrix = D2D1::Matrix3x2F::Translation(
        (logicalSize.Width - (m_geometryBounds.right + m_geometryBounds.left)) / 2,
        (logicalSize.Height - (m_geometryBounds.bottom + m_geometryBounds.top)) / 2);

    context->SetAntialiasMode(D2D1_ANTIALIAS_MODE_ALIASED);
    Matrix3x2F orientationMatrix = m_deviceResources->GetOrientationTransform2D();

    if (m_useSingleFillMesh)
    {
        if (m_textMesh != nullptr)
        {
            context->SetTransform(centerMatrix * orientationMatrix);
            context->FillMesh(m_textMesh.Get(), m_gradientBrush.Get());
        }
    }
    else
    {
        for (TriangleInfo& triangleInfo : m_triangleInfos)
        {
            context->SetTransform(triangleInfo.matrix * 
                                  centerMatrix * orientationMatrix);
            context->FillMesh(triangleInfo.triangleMesh.Get(), 
                              m_gradientBrush.Get());
        }
    }

    // Ignore D2DERR_RECREATE_TARGET here. This error indicates that the device
    // is lost. It will be handled during the next call to Present.
    HRESULT hr = context->EndDraw();
    if (hr != D2DERR_RECREATE_TARGET)
    {
        DX::ThrowIfFailed(hr);
    }

    context->RestoreDrawingState(m_stateBlock.Get());
}

