﻿#include "pch.h"
#include "SparklingTextRenderer.h"
#include "Common/DirectXHelper.h"
#include "InterrogableTessellationSink.h"

using namespace SparklingText;

using namespace D2D1;
using namespace Microsoft::WRL;
using namespace Platform;
using namespace Windows::ApplicationModel;

SparklingTextRenderer::SparklingTextRenderer(const std::shared_ptr<DX::DeviceResources>& deviceResources) : 
    m_text(L""),
    m_useMeshesNotGeometries(false),
    m_deviceResources(deviceResources)
{
    // Create device independent resources
    DX::ThrowIfFailed(
        m_deviceResources->GetDWriteFactory()->CreateTextFormat(
            L"Segoe UI",
            nullptr,
            DWRITE_FONT_WEIGHT_LIGHT,
            DWRITE_FONT_STYLE_NORMAL,
            DWRITE_FONT_STRETCH_NORMAL,
            32.0f,
            L"en-US",
            &m_textFormat)
        );

    DX::ThrowIfFailed(
        m_textFormat->SetParagraphAlignment(DWRITE_PARAGRAPH_ALIGNMENT_FAR)
        );

    DX::ThrowIfFailed(
        m_textFormat->SetTextAlignment(DWRITE_TEXT_ALIGNMENT_TRAILING)
        );

    DX::ThrowIfFailed(
        m_deviceResources->GetD2DFactory()->CreateDrawingStateBlock(&m_stateBlock)
        );

    CreateDeviceDependentResources();
}

void SparklingTextRenderer::CreateDeviceDependentResources()
{
    DX::ThrowIfFailed(
        m_deviceResources->GetD2DDeviceContext()->CreateSolidColorBrush(ColorF(ColorF::White), &m_whiteBrush)
        );

    DX::ThrowIfFailed(
        m_deviceResources->GetD2DDeviceContext()->CreateSolidColorBrush(ColorF(ColorF::Black), &m_solidBrush)
        );
}
void SparklingTextRenderer::ReleaseDeviceDependentResources()
{
    m_whiteBrush.Reset();
    m_solidBrush.Reset();
}

void SparklingTextRenderer::CreateWindowSizeDependentResources()
{
    if (!m_deviceResources->GetLogicalSize().IsEmpty)
    {
        Tessellate();
    }
}

void SparklingTextRenderer::SetRenderMode(bool useMeshesNotGeometries)
{
    m_useMeshesNotGeometries = useMeshesNotGeometries;
    Tessellate();
}

void SparklingTextRenderer::Tessellate()
{
    Windows::Foundation::Size logicalSize = m_deviceResources->GetLogicalSize();

    // Clear both collections
    m_triangleGeometries.clear();
    m_triangleMeshes.clear();

    // Load font file
    ComPtr<IDWriteFactory> dwriteFactory = m_deviceResources->GetDWriteFactory();
    String^ filePath = Package::Current->InstalledLocation->Path + "\\Fonts\\Pesca.ttf";
    ComPtr<IDWriteFontFile> fontFile;

    DX::ThrowIfFailed(
        dwriteFactory->CreateFontFileReference(filePath->Data(), nullptr, &fontFile)
        );

    // Create font face
    ComPtr<IDWriteFontFace> fontFace;

    DX::ThrowIfFailed(
        dwriteFactory->CreateFontFace(DWRITE_FONT_FACE_TYPE_TRUETYPE,
                                      1,
                                      fontFile.GetAddressOf(),
                                      0,
                                      DWRITE_FONT_SIMULATIONS_NONE,
                                      &fontFace)
        );

    // Obtain glyph geometry for "DirectX"
    UINT16 glyphIndices[] = { 0xC3, 0x4C, 0x55, 0x48, 0x46, 0x57, 0x3B };

    ComPtr<ID2D1Factory2> d2dFactory = m_deviceResources->GetD2DFactory();
    ComPtr<ID2D1PathGeometry> pathGeometry;

    DX::ThrowIfFailed(
        d2dFactory->CreatePathGeometry(&pathGeometry)
        );

    ComPtr<ID2D1GeometrySink> geometrySink;

    DX::ThrowIfFailed(
        pathGeometry->Open(&geometrySink)
        );

    DX::ThrowIfFailed(
        fontFace->GetGlyphRunOutline(logicalSize.Width / 4,
                                     glyphIndices,
                                     nullptr,
                                     nullptr,
                                     ARRAYSIZE(glyphIndices),
                                     false,
                                     false,
                                     geometrySink.Get())
        );

    DX::ThrowIfFailed(
        geometrySink->Close()
        );

    // Get the geometry bounds for rendering purposes
    DX::ThrowIfFailed(
        pathGeometry->GetBounds(IdentityMatrix(), &m_geometryBounds)
        );

    // Tessellate geometry into triangles
    ComPtr<InterrogableTessellationSink> tessellationSink = new InterrogableTessellationSink();

    DX::ThrowIfFailed(
        pathGeometry->Tessellate(IdentityMatrix(), tessellationSink.Get())
        );

    std::vector<D2D1_TRIANGLE> triangles = tessellationSink->GetTriangles();

    if (m_useMeshesNotGeometries)
    {
        // Generate a separate mesh from each triangle
        ID2D1DeviceContext* context = m_deviceResources->GetD2DDeviceContext();

        for (D2D1_TRIANGLE triangle : triangles)
        {
            ComPtr<ID2D1Mesh> triangleMesh;

            DX::ThrowIfFailed(
                context->CreateMesh(&triangleMesh)
                );

            ComPtr<ID2D1TessellationSink> sink;

            DX::ThrowIfFailed(
                triangleMesh->Open(&sink)
                );

            sink->AddTriangles(&triangle, 1);
            
            DX::ThrowIfFailed(
                sink->Close()
                );

            m_triangleMeshes.push_back(triangleMesh);
        }
    }
    else
    {
        // Generate a path geometry from each triangle
        for (D2D1_TRIANGLE triangle : triangles)
        {
            ComPtr<ID2D1PathGeometry> triangleGeometry;

            DX::ThrowIfFailed(
                d2dFactory->CreatePathGeometry(&triangleGeometry)
                );

            ComPtr<ID2D1GeometrySink> geometrySink;

            DX::ThrowIfFailed(
                triangleGeometry->Open(&geometrySink)
                );

            geometrySink->BeginFigure(triangle.point1, D2D1_FIGURE_BEGIN_FILLED);
            geometrySink->AddLine(triangle.point2);
            geometrySink->AddLine(triangle.point3);
            geometrySink->EndFigure(D2D1_FIGURE_END_CLOSED);
            
            DX::ThrowIfFailed(
                geometrySink->Close()
                );

            m_triangleGeometries.push_back(triangleGeometry);
        }
    }
}

// Updates the text to be displayed.
void SparklingTextRenderer::Update(DX::StepTimer const& timer)
{
    // Update FPS display text.
    uint32 fps = timer.GetFramesPerSecond();
    m_text = (fps > 0) ? std::to_wstring(fps) + L" FPS" : L" - FPS";
}

// Renders a frame to the screen.
void SparklingTextRenderer::Render()
{
    ID2D1DeviceContext* context = m_deviceResources->GetD2DDeviceContext();
    Windows::Foundation::Size logicalSize = m_deviceResources->GetLogicalSize();

    context->SaveDrawingState(m_stateBlock.Get());
    context->BeginDraw();
    context->Clear(ColorF(0.5f, 0.5f, 0.5f));

    // FPS display
    context->SetTransform(m_deviceResources->GetOrientationTransform2D());
    context->DrawText(m_text.c_str(), 
                      m_text.length(), 
                      m_textFormat.Get(), 
                      RectF(0, 0, logicalSize.Width, logicalSize.Height), 
                      m_whiteBrush.Get());

    // Mesh or Path Geometry display
    Matrix3x2F centerMatrix = D2D1::Matrix3x2F::Translation(
        (logicalSize.Width - (m_geometryBounds.right + m_geometryBounds.left)) / 2,
        (logicalSize.Height - (m_geometryBounds.bottom + m_geometryBounds.top)) / 2);

    context->SetTransform(centerMatrix * m_deviceResources->GetOrientationTransform2D());
    context->SetAntialiasMode(D2D1_ANTIALIAS_MODE_ALIASED);

    if (m_useMeshesNotGeometries)
    {
        for (ComPtr<ID2D1Mesh>& triangleMesh : m_triangleMeshes)
        {
            float gray = (rand() % 1000) * 0.001f;
            m_solidBrush->SetColor(ColorF(gray, gray, gray));
            context->FillMesh(triangleMesh.Get(), m_solidBrush.Get());
        }
    }
    else
    {
        for (ComPtr<ID2D1PathGeometry>& triangleGeometry : m_triangleGeometries)
        {
            float gray = (rand() % 1000) * 0.001f;
            m_solidBrush->SetColor(ColorF(gray, gray, gray));
            context->FillGeometry(triangleGeometry.Get(), m_solidBrush.Get());
        }
    }

    // Ignore D2DERR_RECREATE_TARGET here. This error indicates that the device
    // is lost. It will be handled during the next call to Present.
    HRESULT hr = context->EndDraw();
    if (hr != D2DERR_RECREATE_TARGET)
    {
        DX::ThrowIfFailed(hr);
    }

    context->RestoreDrawingState(m_stateBlock.Get());
}

