﻿#include "pch.h"
#include "TessellationVisualizationRenderer.h"
#include "Common/DirectXHelper.h"
#include "InterrogableTessellationSink.h"

using namespace TessellationVisualization;

using namespace D2D1;
using namespace Microsoft::WRL;
using namespace Platform;
using namespace Windows::ApplicationModel;

TessellationVisualizationRenderer::TessellationVisualizationRenderer(const std::shared_ptr<DX::DeviceResources>& deviceResources) : 
    m_needsRender(true),
    m_geometryType(GeometryType::None),
    m_renderOption(RenderOption::Outline),
    m_deviceResources(deviceResources)
{
    ComPtr<ID2D1Factory2> d2dFactory = m_deviceResources->GetD2DFactory();

    DX::ThrowIfFailed(
        d2dFactory->CreateDrawingStateBlock(&m_stateBlock)
        );

    DX::ThrowIfFailed(
        d2dFactory->CreateStrokeStyle(StrokeStyleProperties(D2D1_CAP_STYLE_FLAT, 
                                                            D2D1_CAP_STYLE_FLAT, 
                                                            D2D1_CAP_STYLE_FLAT, 
                                                            D2D1_LINE_JOIN_BEVEL), 
                                      nullptr, 0, &m_strokeStyle)
        );

    CreateDeviceDependentResources();
}

void TessellationVisualizationRenderer::CreateDeviceDependentResources()
{
    ComPtr<ID2D1DeviceContext1> context = m_deviceResources->GetD2DDeviceContext();

    DX::ThrowIfFailed(
        context->CreateSolidColorBrush(ColorF(ColorF::Black), &m_blackBrush)
        );

    DX::ThrowIfFailed(
        context->CreateSolidColorBrush(ColorF(ColorF::Black), &m_fillBrush)
        );

    m_needsRender = true;
}
void TessellationVisualizationRenderer::ReleaseDeviceDependentResources()
{
    m_blackBrush.Reset();
    m_fillBrush.Reset();
}

void TessellationVisualizationRenderer::CreateWindowSizeDependentResources()
{
    GenerateGeometries();
    m_needsRender = true;
}

// Program options
void TessellationVisualizationRenderer::SetGeometryType(GeometryType geometryType)
{
    m_geometryType = geometryType;
    GenerateGeometries();
    m_needsRender = true;
}

void TessellationVisualizationRenderer::SetRenderOption(RenderOption renderOption)
{
    m_renderOption = renderOption;
    m_needsRender = true;
}

void TessellationVisualizationRenderer::GenerateGeometries()
{
    // Clear current collection
    m_pathGeometries.clear();

    Windows::Foundation::Size logicalSize = m_deviceResources->GetLogicalSize();

    if (logicalSize.IsEmpty)
        return;

    float margin = 50;
    float dimension = min(logicalSize.Width, logicalSize.Height) - 2 * margin;

    ComPtr<ID2D1Factory2> d2dFactory = m_deviceResources->GetD2DFactory();
    ComPtr<ID2D1Geometry> geometry = nullptr;

    switch (m_geometryType)
    {
    case GeometryType::None:
        break;

    case GeometryType::Rectangle:
        {
            ComPtr<ID2D1RectangleGeometry> rectangleGeometry;

            DX::ThrowIfFailed(
                d2dFactory->CreateRectangleGeometry(RectF(0, 0, dimension, dimension), &rectangleGeometry)
                );

            rectangleGeometry.As(&geometry);
        }
        break; 

    case GeometryType::RoundedRectangle:
        {
            ComPtr<ID2D1RoundedRectangleGeometry> roundedRectangleGeometry;

            DX::ThrowIfFailed(
                d2dFactory->CreateRoundedRectangleGeometry(RoundedRect(RectF(0, 0, dimension, dimension),
                                                                       dimension / 8, dimension / 8), 
                                                           &roundedRectangleGeometry)
                );

            roundedRectangleGeometry.As(&geometry);
        }
        break;

    case GeometryType::Ellipse:
        {
            ComPtr<ID2D1EllipseGeometry> ellipseGeometry;

            DX::ThrowIfFailed(
                d2dFactory->CreateEllipseGeometry(Ellipse(Point2F(),
                                                          dimension / 2, dimension / 2), 
                                                  &ellipseGeometry)
                );

            ellipseGeometry.As(&geometry);
        }
        break;

    case GeometryType::FilledInfinitySign:
        {
            float radius = dimension / 5;
            float x1 = 0.50f * radius;
            float x2 = 0.95f * radius;
            float x3 = 1.50f * radius;
            float x4 = 2.05f * radius;
            float x5 = 2.50f * radius;

            float y1 = 0.50f * radius;
            float y2 = 0.55f * radius;
            float y3 = 1.00f * radius;

            ComPtr<ID2D1PathGeometry> pathGeometry;

            DX::ThrowIfFailed(
                d2dFactory->CreatePathGeometry(&pathGeometry)
                );

            ComPtr<ID2D1GeometrySink> geometrySink;

            DX::ThrowIfFailed(
                pathGeometry->Open(&geometrySink)
                );

            geometrySink->BeginFigure(Point2F(-x3, -y3), D2D1_FIGURE_BEGIN_FILLED);
            geometrySink->AddBezier(BezierSegment(Point2F(-x4, -y3), Point2F(-x5, -y2), Point2F(-x5, 0)));
            geometrySink->AddBezier(BezierSegment(Point2F(-x5, y2), Point2F(-x4, y3), Point2F(-x3, y3)));
            geometrySink->AddBezier(BezierSegment(Point2F(-x2, y3), Point2F(-x1, y1), Point2F(0, 0)));
            geometrySink->AddBezier(BezierSegment(Point2F(x1, -y1), Point2F(x2, -y3), Point2F(x3, -y3)));
            geometrySink->AddBezier(BezierSegment(Point2F(x4, -y3), Point2F(x5, -y2), Point2F(x5, 0)));
            geometrySink->AddBezier(BezierSegment(Point2F(x5, y2), Point2F(x4, y3), Point2F(x3, y3)));
            geometrySink->AddBezier(BezierSegment(Point2F(x2, y3), Point2F(x1, y1), Point2F(0, 0)));
            geometrySink->AddBezier(BezierSegment(Point2F(-x1, -y1), Point2F(-x2, -y3), Point2F(-x3, -y3)));
            geometrySink->EndFigure(D2D1_FIGURE_END_CLOSED);

            DX::ThrowIfFailed(
                geometrySink->Close()
                );

            pathGeometry.As(&geometry);
        }
        break;

    case GeometryType::WidenedBezierStroke:
    case GeometryType::DottedBezierStroke:
        {
            ComPtr<ID2D1PathGeometry> pathGeometry;

            DX::ThrowIfFailed(
                d2dFactory->CreatePathGeometry(&pathGeometry)
                );

            ComPtr<ID2D1GeometrySink> geometrySink;

            DX::ThrowIfFailed(
                pathGeometry->Open(&geometrySink)
                );

            geometrySink->BeginFigure(Point2F(2 * logicalSize.Width / 5, 0), D2D1_FIGURE_BEGIN_HOLLOW);
            geometrySink->AddBezier(BezierSegment(Point2F(logicalSize.Width, logicalSize.Height),
                                                  Point2F(0, logicalSize.Height),
                                                  Point2F(3 * logicalSize.Width / 5, 0)));
            
            geometrySink->EndFigure(D2D1_FIGURE_END_OPEN);

            DX::ThrowIfFailed(
                geometrySink->Close()
                );

            ComPtr<ID2D1PathGeometry> widenedPathGeometry;

            DX::ThrowIfFailed(
                d2dFactory->CreatePathGeometry(&widenedPathGeometry)
                );

            DX::ThrowIfFailed(
                widenedPathGeometry->Open(&geometrySink)
                );

            ComPtr<ID2D1StrokeStyle> strokeStyle;
            bool isDotted = m_geometryType == GeometryType::DottedBezierStroke;
            float dashes[] = { 0, 2 };

            DX::ThrowIfFailed(
                d2dFactory->CreateStrokeStyle(StrokeStyleProperties(D2D1_CAP_STYLE_ROUND,
                                                                    D2D1_CAP_STYLE_ROUND, 
                                                                    D2D1_CAP_STYLE_ROUND, 
                                                                    D2D1_LINE_JOIN_MITER, 
                                                                    10.0f, 
                                                                    isDotted ? D2D1_DASH_STYLE_CUSTOM : D2D1_DASH_STYLE_SOLID),
                                              isDotted ? dashes : 0,
                                              isDotted ? 2 : 0, 
                                              &strokeStyle)
                );

            DX::ThrowIfFailed(
                pathGeometry->Widen(25, strokeStyle.Get(), IdentityMatrix(), geometrySink.Get())
                );

            DX::ThrowIfFailed(
                geometrySink->Close()
                );

            widenedPathGeometry.As(&geometry);
        }
        break;

    case GeometryType::SansSerifText:
    case GeometryType::SerifText:
        {
            bool isSansSerif = m_geometryType == GeometryType::SansSerifText;

            // Load font file
            ComPtr<IDWriteFactory> dwriteFactory = m_deviceResources->GetDWriteFactory();
            String^ filePath = Package::Current->InstalledLocation->Path + "\\Fonts\\" +
                               (isSansSerif ? "Miramo" : "Pesca")  + ".ttf";
            ComPtr<IDWriteFontFile> fontFile;

            DX::ThrowIfFailed(
                dwriteFactory->CreateFontFileReference(filePath->Data(), nullptr, &fontFile)
                );

            // Create font face
            ComPtr<IDWriteFontFace> fontFace;

            DX::ThrowIfFailed(
                dwriteFactory->CreateFontFace(DWRITE_FONT_FACE_TYPE_TRUETYPE,
                                              1,
                                              fontFile.GetAddressOf(),
                                              0,
                                              DWRITE_FONT_SIMULATIONS_NONE,
                                              &fontFace)
                );

            // Get glyph indices
            UINT32 codePoints [] = { 'D', 'X' };
            UINT16 glyphIndices[2];

            DX::ThrowIfFailed(
                fontFace->GetGlyphIndices(codePoints, ARRAYSIZE(codePoints), glyphIndices)
                );

            // Obtain glyph geometry 
            ComPtr<ID2D1PathGeometry> pathGeometry;

            DX::ThrowIfFailed(
                d2dFactory->CreatePathGeometry(&pathGeometry)
                );

            ComPtr<ID2D1GeometrySink> geometrySink;

            DX::ThrowIfFailed(
                pathGeometry->Open(&geometrySink)
                );

            DX::ThrowIfFailed(
                fontFace->GetGlyphRunOutline(dimension,
                                             glyphIndices,
                                             nullptr,
                                             nullptr,
                                             ARRAYSIZE(glyphIndices),
                                             false,
                                             false,
                                             geometrySink.Get())
                );

            DX::ThrowIfFailed(
                geometrySink->Close()
                );

            pathGeometry.As(&geometry);
        }
        break;
    }

    if (geometry == nullptr)
        return;

    // Get the geometry bounds for rendering purposes
    DX::ThrowIfFailed(
        geometry->GetBounds(IdentityMatrix(), &m_geometryBounds)
        );

    // Tessellate geometry into triangles
    ComPtr<InterrogableTessellationSink> tessellationSink = new InterrogableTessellationSink();
    geometry->Tessellate(IdentityMatrix(), tessellationSink.Get());
    std::vector<D2D1_TRIANGLE> triangles = tessellationSink->GetTriangles();

    // Generate a path geometry from each triangle
    for (D2D1_TRIANGLE triangle : triangles)
    {
        ComPtr<ID2D1PathGeometry> pathGeometry;

        DX::ThrowIfFailed(
            d2dFactory->CreatePathGeometry(&pathGeometry)
            );

        ComPtr<ID2D1GeometrySink> geometrySink;

        DX::ThrowIfFailed(
            pathGeometry->Open(&geometrySink)
            );

        geometrySink->BeginFigure(triangle.point1, D2D1_FIGURE_BEGIN_FILLED);
        geometrySink->AddLine(triangle.point2);
        geometrySink->AddLine(triangle.point3);
        geometrySink->EndFigure(D2D1_FIGURE_END_CLOSED);

        DX::ThrowIfFailed(
            geometrySink->Close()
            );

        m_pathGeometries.push_back(pathGeometry);
    }
}

void TessellationVisualizationRenderer::Update(DX::StepTimer const& timer)
{
}

// Renders a frame to the screen.
bool TessellationVisualizationRenderer::Render()
{
    if (!m_needsRender)
        return false;

    ID2D1DeviceContext* context = m_deviceResources->GetD2DDeviceContext();
    Windows::Foundation::Size logicalSize = m_deviceResources->GetLogicalSize();

    context->SaveDrawingState(m_stateBlock.Get());
    context->BeginDraw();
    context->Clear(ColorF(ColorF::LightYellow));

    Matrix3x2F centerMatrix = D2D1::Matrix3x2F::Translation(
        (logicalSize.Width - (m_geometryBounds.right + m_geometryBounds.left)) / 2,
        (logicalSize.Height - (m_geometryBounds.bottom + m_geometryBounds.top)) / 2);

    context->SetTransform(centerMatrix * m_deviceResources->GetOrientationTransform2D());

    switch (m_renderOption)
    {
    case RenderOption::Outline:
        for (ComPtr<ID2D1PathGeometry>& pathGeometry : m_pathGeometries)
        {
            context->DrawGeometry(pathGeometry.Get(), m_blackBrush.Get(), 1.0f, m_strokeStyle.Get());
        }
        break;

    case RenderOption::GradientFill:
        context->SetAntialiasMode(D2D1_ANTIALIAS_MODE_ALIASED);

        unsigned int size = m_pathGeometries.size();

        for (unsigned int index = 0; index < size; index++)
        {
            ComPtr<ID2D1PathGeometry>& pathGeometry = m_pathGeometries.at(index);

            ColorF color = ColorF(float(index) / size, 0, float(size - index) / size);
            m_fillBrush->SetColor(color);

            context->FillGeometry(pathGeometry.Get(), m_fillBrush.Get());
        }
        break;
    }

    // Ignore D2DERR_RECREATE_TARGET here. This error indicates that the device
    // is lost. It will be handled during the next call to Present.
    HRESULT hr = context->EndDraw();
    if (hr != D2DERR_RECREATE_TARGET)
    {
        DX::ThrowIfFailed(hr);
    }

    context->RestoreDrawingState(m_stateBlock.Get());

    m_needsRender = false;
    return true;
}

