﻿#include "pch.h"
#include "TextMorphingRenderer.h"
#include "Common/DirectXHelper.h"
#include "InterrogableTessellationSink.h"

using namespace TextMorphing;

using namespace D2D1;
using namespace Microsoft::WRL;
using namespace Platform;
using namespace Windows::ApplicationModel;

TextMorphingRenderer::TextMorphingRenderer(const std::shared_ptr<DX::DeviceResources>& deviceResources) : 
    m_text(L""),
    m_textMesh(nullptr),
    m_deviceResources(deviceResources)
{
    // Create device independent resources
    DX::ThrowIfFailed(
        m_deviceResources->GetDWriteFactory()->CreateTextFormat(
            L"Segoe UI",
            nullptr,
            DWRITE_FONT_WEIGHT_LIGHT,
            DWRITE_FONT_STYLE_NORMAL,
            DWRITE_FONT_STRETCH_NORMAL,
            32.0f,
            L"en-US",
            &m_textFormat
            )
        );

    DX::ThrowIfFailed(
        m_textFormat->SetParagraphAlignment(DWRITE_PARAGRAPH_ALIGNMENT_FAR)
        );

    DX::ThrowIfFailed(
        m_textFormat->SetTextAlignment(DWRITE_TEXT_ALIGNMENT_TRAILING)
        );

    DX::ThrowIfFailed(
        m_deviceResources->GetD2DFactory()->CreateDrawingStateBlock(&m_stateBlock)
        );

    CreateDeviceDependentResources();
}

void TextMorphingRenderer::CreateDeviceDependentResources()
{
    DX::ThrowIfFailed(
        m_deviceResources->GetD2DDeviceContext()->CreateSolidColorBrush(ColorF(ColorF::Blue), &m_blueBrush)
        );
}
void TextMorphingRenderer::ReleaseDeviceDependentResources()
{
    m_blueBrush.Reset();
}

void TextMorphingRenderer::CreateWindowSizeDependentResources()
{
    Windows::Foundation::Size logicalSize = m_deviceResources->GetLogicalSize();

    if (logicalSize.IsEmpty)
        return;

    // Load font file
    ComPtr<IDWriteFactory> dwriteFactory = m_deviceResources->GetDWriteFactory();
    String^ filePath = Package::Current->InstalledLocation->Path + "\\Fonts\\Pesca.ttf";
    ComPtr<IDWriteFontFile> fontFile;

    DX::ThrowIfFailed(
        dwriteFactory->CreateFontFileReference(filePath->Data(), nullptr, &fontFile)
        );

    // Create font face
    ComPtr<IDWriteFontFace> fontFace;

    DX::ThrowIfFailed(
        dwriteFactory->CreateFontFace(DWRITE_FONT_FACE_TYPE_TRUETYPE,
                                      1,
                                      fontFile.GetAddressOf(),
                                      0,
                                      DWRITE_FONT_SIMULATIONS_NONE,
                                      &fontFace)
        );

    ComPtr<ID2D1Factory2> d2dFactory = m_deviceResources->GetD2DFactory();
    D2D1_RECT_F geometryBounds[2];
    std::vector<D2D1_TRIANGLE> triangles[2];

    // Do similar logic for the words "DirectX" and "Factor"
    for (int word = 0; word < 2; word++)
    {
        // Glyph indices for "DirectX" and "Factor"
        UINT16 glyphIndicesDirectX [] = { 0xC3, 0x4C, 0x55, 0x48, 0x46, 0x57, 0x3B };
        UINT16 glyphIndicesFactor [] = { 0xC5, 0x44, 0x46, 0x57, 0x52, 0x55 };

        // Get glyph outline geometry
        ComPtr<ID2D1PathGeometry> pathGeometry;

        DX::ThrowIfFailed(
            d2dFactory->CreatePathGeometry(&pathGeometry)
            );

        ComPtr<ID2D1GeometrySink> geometrySink;

        DX::ThrowIfFailed(
            pathGeometry->Open(&geometrySink)
            );

        DX::ThrowIfFailed(
            fontFace->GetGlyphRunOutline(0.3f * logicalSize.Width,
                                         word == 0 ? glyphIndicesDirectX : glyphIndicesFactor,
                                         nullptr,
                                         nullptr,
                                         word == 0 ? ARRAYSIZE(glyphIndicesDirectX) : ARRAYSIZE(glyphIndicesFactor),
                                         false,
                                         false,
                                         geometrySink.Get())
            );

        DX::ThrowIfFailed(
            geometrySink->Close()
            );

        // Get the geometry bounds for rendering purposes
        DX::ThrowIfFailed(
            pathGeometry->GetBounds(IdentityMatrix(), &geometryBounds[word])
            );

        // Tessellate geometry into triangles
        ComPtr<InterrogableTessellationSink> tessellationSink = new InterrogableTessellationSink();

        DX::ThrowIfFailed(
            pathGeometry->Tessellate(IdentityMatrix(), tessellationSink.Get())
            );

        triangles[word] = tessellationSink->GetTriangles();
    }

    // Calculate composite bounding rect
    m_geometryBounds = RectF(
        min(geometryBounds[0].left, geometryBounds[1].left),
        min(geometryBounds[0].top, geometryBounds[1].top),
        max(geometryBounds[0].right, geometryBounds[1].right),
        max(geometryBounds[0].bottom, geometryBounds[1].bottom));

    // Two functions for making adjustments for different bounding rects
    std::function<void(D2D1_POINT_2F&, float, float)> AdjustPoint =
        [] (D2D1_POINT_2F& pt, float adjX, float adjY)
    {
        pt.x += adjX;
        pt.y += adjY;
    };

    std::function<void(D2D1_TRIANGLE&, float, float)> AdjustTriangle =
        [AdjustPoint] (D2D1_TRIANGLE& triangle, float adjX, float adjY)
    {
        AdjustPoint(triangle.point1, adjX, adjY);
        AdjustPoint(triangle.point2, adjX, adjY);
        AdjustPoint(triangle.point3, adjX, adjY);
    };

    // Adjust triangle coordinates for different bounding boxes
    float compositeWidth = m_geometryBounds.right - m_geometryBounds.left;
    float compositeHeight = m_geometryBounds.bottom - m_geometryBounds.top;

    for (int word = 0; word < 2; word++)
    {
        float width = geometryBounds[word].right - geometryBounds[word].left;
        float height = geometryBounds[word].bottom - geometryBounds[word].top;

        float adjX = (compositeWidth - width) / 2;
        float adjY = (compositeHeight - height) / 2;

        for (D2D1_TRIANGLE& triangle : triangles[word])
        {
            AdjustTriangle(triangle, adjX, adjY);
        }
    }

    // Get maximum triangle count
    int count = max(triangles[0].size(), triangles[1].size());

    // Sort both triangle collections by x coordinate
    for (int word = 0; word < 2; word++)
    {
        std::sort(triangles[word].begin(), 
                  triangles[word].end(), 
                  [] (D2D1_TRIANGLE tri1, D2D1_TRIANGLE tri2)
        {
            return (tri1.point1.x + tri1.point2.x + tri1.point3.x) <
                   (tri2.point1.x + tri2.point2.x + tri2.point3.x);
        });
    }

    // Sort within groups from left to right by y coordinate 
    int groupCount = (int)sqrt(count);

    for (int begin = 0; begin < count; begin += groupCount)
    {
        int end = min(count, begin + groupCount);
        int begins[2];
        int ends[2];

        for (int word = 0; word < 2; word++)
        {
            begins[word] = begin * triangles[word].size() / count;
            ends[word] = end * triangles[word].size() / count;

            std::sort(triangles[word].begin() + begins[word], 
                      triangles[word].begin() + ends[word], 
                      [] (D2D1_TRIANGLE tri1, D2D1_TRIANGLE tri2)
            {
                return (tri1.point1.y + tri1.point2.y + tri1.point3.y) <
                       (tri2.point1.y + tri2.point2.y + tri2.point3.y);
            });
        }
    }

    // Assemble TriangleInfo objects
    m_triangleInfos = std::vector<TriangleInfo>(count);

    for (int index = 0; index < count; index++)
    {
        int index0 = index * triangles[0].size() / count;
        int index1 = index * triangles[1].size() / count;

        TriangleInfo triangleInfo;
        triangleInfo.triangle[0] = triangles[0].at(index0);
        triangleInfo.triangle[1] = triangles[1].at(index1);

        m_triangleInfos.at(index) = triangleInfo;
    }
}

void TextMorphingRenderer::Update(DX::StepTimer const& timer)
{
    // Update FPS display text.
    uint32 fps = timer.GetFramesPerSecond();
    m_text = (fps > 0) ? std::to_wstring(fps) + L" FPS" : L" - FPS";

    if (m_triangleInfos.size() == 0)
        return;

    // Calculate an interpolation factor
    float t = (float)fmod(timer.GetTotalSeconds(), 10) / 10;
    t = std::cos(t * 2 * 3.14159f);     // 1 to 0 to -1 to 0 to 1
    t = (1 - t) / 2;                    // 0 to 1 to 0

    // Two functions for interpolation
    std::function<D2D1_POINT_2F(D2D1_POINT_2F, D2D1_POINT_2F, float)> InterpolatePoint =
        [](D2D1_POINT_2F pt0, D2D1_POINT_2F pt1, float t)
    {
        return Point2F((1 - t) * pt0.x + t * pt1.x,
            (1 - t) * pt0.y + t * pt1.y);
    };

    std::function<D2D1_TRIANGLE(D2D1_TRIANGLE, D2D1_TRIANGLE, float)> InterpolateTriangle =
        [InterpolatePoint](D2D1_TRIANGLE tri0, D2D1_TRIANGLE tri1, float t)
    {
        D2D1_TRIANGLE triangle;
        triangle.point1 = InterpolatePoint(tri0.point1, tri1.point1, t);
        triangle.point2 = InterpolatePoint(tri0.point2, tri1.point2, t);
        triangle.point3 = InterpolatePoint(tri0.point3, tri1.point3, t);
        return triangle;
    };

    // Interpolate the triangles
    int count = m_triangleInfos.size();
    std::vector<D2D1_TRIANGLE> triangles(count);

    for (int index = 0; index < count; index++)
    {
        triangles.at(index) = InterpolateTriangle(m_triangleInfos.at(index).triangle[0], 
                                                  m_triangleInfos.at(index).triangle[1], t);
    }

    // Create a mesh with the interpolated triangles
    DX::ThrowIfFailed(
        m_deviceResources->GetD2DDeviceContext()->CreateMesh(&m_textMesh)
        );

    ComPtr<ID2D1TessellationSink> tessellationSink;

    DX::ThrowIfFailed(
        m_textMesh->Open(&tessellationSink)
        );

    tessellationSink->AddTriangles(triangles.data(), triangles.size());

    DX::ThrowIfFailed(
        tessellationSink->Close()
        );
}

// Renders a frame to the screen.
void TextMorphingRenderer::Render()
{
    ID2D1DeviceContext* context = m_deviceResources->GetD2DDeviceContext();
    Windows::Foundation::Size logicalSize = m_deviceResources->GetLogicalSize();

    context->SaveDrawingState(m_stateBlock.Get());
    context->BeginDraw();
    context->Clear(ColorF(ColorF::MistyRose));

    // Display the FPS text
    context->SetTransform(m_deviceResources->GetOrientationTransform2D());
    context->DrawText(m_text.c_str(), 
                      m_text.length(), 
                      m_textFormat.Get(), 
                      RectF(0, 0, logicalSize.Width, logicalSize.Height), 
                      m_blueBrush.Get());

    if (m_textMesh != nullptr)
    {
        // Mesh display
        Matrix3x2F centerMatrix = D2D1::Matrix3x2F::Translation(
            (logicalSize.Width - (m_geometryBounds.right + m_geometryBounds.left)) / 2,
            (logicalSize.Height - (m_geometryBounds.bottom + m_geometryBounds.top)) / 2);

        context->SetTransform(centerMatrix * m_deviceResources->GetOrientationTransform2D());
        context->SetAntialiasMode(D2D1_ANTIALIAS_MODE_ALIASED);
        context->FillMesh(m_textMesh.Get(), m_blueBrush.Get());
    }

    // Ignore D2DERR_RECREATE_TARGET here. This error indicates that the device
    // is lost. It will be handled during the next call to Present.
    HRESULT hr = context->EndDraw();
    if (hr != D2DERR_RECREATE_TARGET)
    {
        DX::ThrowIfFailed(hr);
    }

    context->RestoreDrawingState(m_stateBlock.Get());
}

